use crate::ctx::Context;
use crate::dbs::{Options, Transaction};
use crate::err::Error;
use crate::iam::{Action, ResourceKind};
use crate::sql::{Base, Ident, Value};
use derive::Store;
use revision::revisioned;
use serde::{Deserialize, Serialize};
use std::fmt::{self, Display, Formatter};

#[revisioned(revision = 2)]
#[derive(Clone, Debug, Default, Eq, PartialEq, PartialOrd, Serialize, Deserialize, Store, Hash)]
#[cfg_attr(feature = "arbitrary", derive(arbitrary::Arbitrary))]
pub struct RemoveAnalyzerStatement {
	pub name: Ident,
	#[revision(start = 2)]
	pub if_exists: bool,
}

impl RemoveAnalyzerStatement {
	pub(crate) async fn compute(
		&self,
		_ctx: &Context<'_>,
		opt: &Options,
		txn: &Transaction,
	) -> Result<Value, Error> {
		let r = async {
			// Allowed to run?
			opt.is_allowed(Action::Edit, ResourceKind::Analyzer, &Base::Db)?;
			// Claim transaction
			let mut run = txn.lock().await;
			// Clear the cache
			run.clear_cache();
			// Get the definition
			let az = run.get_db_analyzer(opt.ns(), opt.db(), &self.name).await?;
			// Delete the definition
			let key = crate::key::database::az::new(opt.ns(), opt.db(), &az.name);
			run.del(key).await?;
			// TODO Check that the analyzer is not used in any schema
			// Ok all good
			Ok(Value::None)
		}
		.await;
		match r {
			Err(Error::AzNotFound {
				..
			}) if self.if_exists => Ok(Value::None),
			v => v,
		}
	}
}

impl Display for RemoveAnalyzerStatement {
	fn fmt(&self, f: &mut Formatter) -> fmt::Result {
		write!(f, "REMOVE ANALYZER")?;
		if self.if_exists {
			write!(f, " IF EXISTS")?
		}
		write!(f, " {}", self.name)?;
		Ok(())
	}
}
