use super::super::super::{
	comment::shouldbespace,
	ending,
	error::expected,
	literal::{ident, strand},
	part::permission::permission,
	value::value,
	IResult,
};
use crate::{
	sql::{statements::DefineParamStatement, Permission, Strand, Value},
	syn::v1::ParseError,
};
use nom::{
	branch::alt, bytes::complete::tag_no_case, character::complete::char, combinator::cut,
	multi::many0, Err,
};
use nom::{combinator::opt, sequence::tuple};

pub fn param(i: &str) -> IResult<&str, DefineParamStatement> {
	let (i, _) = tag_no_case("PARAM")(i)?;
	let (i, if_not_exists) = opt(tuple((
		shouldbespace,
		tag_no_case("IF"),
		shouldbespace,
		tag_no_case("NOT"),
		shouldbespace,
		cut(tag_no_case("EXISTS")),
	)))(i)?;
	let (i, _) = shouldbespace(i)?;
	let (i, _) = cut(char('$'))(i)?;
	let (i, name) = cut(ident)(i)?;
	let (i, opts) = many0(param_opts)(i)?;
	let (i, _) = expected("VALUE, PERMISSIONS, or COMMENT", ending::query)(i)?;
	// Create the base statement
	let mut res = DefineParamStatement {
		name,
		if_not_exists: if_not_exists.is_some(),
		..Default::default()
	};
	// Assign any defined options
	for opt in opts {
		match opt {
			DefineParamOption::Value(v) => {
				res.value = v;
			}
			DefineParamOption::Comment(v) => {
				res.comment = Some(v);
			}
			DefineParamOption::Permissions(v) => {
				res.permissions = v;
			}
		}
	}
	// Check necessary options
	if res.value.is_none() {
		return Err(Err::Failure(ParseError::ExplainedExpected {
			tried: i,
			expected: "a VALUE clause",
			explained: "A param requires a VALUE clause to be defined.",
		}));
	}
	// Return the statement
	Ok((i, res))
}

enum DefineParamOption {
	Value(Value),
	Comment(Strand),
	Permissions(Permission),
}

fn param_opts(i: &str) -> IResult<&str, DefineParamOption> {
	alt((param_value, param_comment, param_permissions))(i)
}

fn param_value(i: &str) -> IResult<&str, DefineParamOption> {
	let (i, _) = shouldbespace(i)?;
	let (i, _) = tag_no_case("VALUE")(i)?;
	let (i, _) = shouldbespace(i)?;
	let (i, v) = cut(value)(i)?;
	Ok((i, DefineParamOption::Value(v)))
}

fn param_comment(i: &str) -> IResult<&str, DefineParamOption> {
	let (i, _) = shouldbespace(i)?;
	let (i, _) = tag_no_case("COMMENT")(i)?;
	let (i, _) = shouldbespace(i)?;
	let (i, v) = cut(strand)(i)?;
	Ok((i, DefineParamOption::Comment(v)))
}

fn param_permissions(i: &str) -> IResult<&str, DefineParamOption> {
	let (i, _) = shouldbespace(i)?;
	let (i, _) = tag_no_case("PERMISSIONS")(i)?;
	let (i, _) = shouldbespace(i)?;
	let (i, v) = cut(permission)(i)?;
	Ok((i, DefineParamOption::Permissions(v)))
}

#[cfg(test)]
mod tests {

	use super::*;

	#[test]
	fn define_param_without_value_clause() {
		let sql = "PARAM test";
		let res = param(sql);

		assert!(res.is_err())
	}
}
